import { UUID } from '@/types'

export interface WorkerConfiguration {
  id: UUID
  name: string
  configuration: Record<string, unknown>
  archived: boolean
}

/**
 * Type names that are allowed as the `subtype` of a list
 */
export interface UserConfigurationListSubtypes {
  string: string
  int: number
  float: number
  bool: boolean
}

interface _UserConfigurationField<TypeName extends string, ValueType> {
  type: TypeName
  title: string
  required?: boolean
  default?: ValueType
}

/*
 * A list configuration field gets an extra `subtype` property,
 * and the field value's type depends on that subtype, since it will be an array of that subtype.
 * So for each list subtype, we create a distinct user configuration field.
 */
type ListUserConfigurationFields = {
  [Subtype in keyof UserConfigurationListSubtypes]: _UserConfigurationField<'list', UserConfigurationListSubtypes[Subtype][]> & { subtype: Subtype }
}

export type ListUserConfigurationField = ListUserConfigurationFields[keyof ListUserConfigurationFields]

/**
 * For each subtype name of UserConfigurationListSubtypes, build an array of the corresponding TypeScript type
 */
type ListUserConfigurationFieldValues = { [Subtype in keyof UserConfigurationListSubtypes]: UserConfigurationListSubtypes[Subtype][] }

/**
 * All the possible types for the values of a ListField, of any subtype
 */
export type ListUserConfigurationFieldValue = ListUserConfigurationFieldValues[keyof ListUserConfigurationFieldValues]

export interface EnumUserConfigurationField extends _UserConfigurationField<'enum', string> {
  choices: string[]
}

export type StringUserConfigurationField = _UserConfigurationField<'string', string>

export type IntegerUserConfigurationField = _UserConfigurationField<'int', number>

export type FloatUserConfigurationField = _UserConfigurationField<'float', number>

export type BooleanUserConfigurationField = _UserConfigurationField<'bool', boolean>

export type DictUserConfigurationField = _UserConfigurationField<'dict', Record<string, string>>

export type ModelUserConfigurationField = _UserConfigurationField<'model', UUID>

export interface UserConfigurationFields {
  string: StringUserConfigurationField
  int: IntegerUserConfigurationField
  float: FloatUserConfigurationField
  bool: BooleanUserConfigurationField
  dict: DictUserConfigurationField
  enum: EnumUserConfigurationField
  list: ListUserConfigurationField
  model: ModelUserConfigurationField
}

export type UserConfigurationField = UserConfigurationFields[keyof UserConfigurationFields]
