import axios from 'axios'
import { assert } from 'chai'
import { createPinia, setActivePinia } from 'pinia'
import { initialState, mutations, getters } from '@/store/selection'
import { useJobsStore } from '@/stores'
import store from './index.spec.js'
import { assertRejects, FakeAxios } from '../testhelpers'
import { jobsSample } from '../samples'

describe('selection', () => {
  describe('mutations', () => {
    it('set', () => {
      const state = {
        count: null,
        selection: {
          corpus1: ['element1']
        }
      }
      mutations.set(state, {
        count: 4,
        results: [
          { id: 'element1', corpus: { id: 'corpus1' } },
          { id: 'element2', corpus: { id: 'corpus1' } },
          { id: 'element3', corpus: { id: 'corpus2' } }
        ]
      })
      assert.deepStrictEqual(state, {
        count: 4,
        selection: {
          corpus1: ['element1', 'element2'],
          corpus2: ['element3']
        }
      })
    })
    it('remove', () => {
      const state = {
        count: 3,
        selection: {
          corpus1: ['element1', 'element2'],
          corpus2: ['element3']
        }
      }
      mutations.remove(state, { id: 'element2', corpus: { id: 'corpus1' } })
      assert.deepStrictEqual(state, {
        count: 2,
        selection: {
          corpus1: ['element1'],
          corpus2: ['element3']
        }
      })
    })
    it('removeCorpus', () => {
      const state = {
        count: 3,
        selection: {
          corpus1: ['element1', 'element2'],
          corpus2: ['element3']
        }
      }
      mutations.removeCorpus(state, 'corpus1')
      assert.deepStrictEqual(state, {
        count: 1,
        selection: {
          corpus2: ['element3']
        }
      })
    })
    it('reset', () => {
      const state = {
        count: 3,
        selection: {
          corpus1: ['element1', 'element2'],
          corpus2: ['element3']
        }
      }
      mutations.reset(state)
      assert.deepStrictEqual(state, initialState())
    })
  })

  describe('actions', () => {
    let mock, jobsStore

    before('Setting up mocks', () => {
      mock = new FakeAxios(axios)
      setActivePinia(createPinia())
      jobsStore = useJobsStore()
    })

    afterEach(() => {
      // Remove any handlers, but leave mocking in place
      mock.reset()
      store.reset()
      jobsStore.$reset()
    })

    after('Removing Axios mock', () => {
      // Remove mocking entirely
      mock.restore()
    })

    describe('get', () => {
      it('goes through the pagination', async () => {
        const element1 = { id: 'element1', name: 'Element 1', corpus: { id: 'corpus1' } }
        const element2 = { id: 'element2', name: 'Element 2', corpus: { id: 'corpus1' } }
        const element3 = { id: 'element3', name: 'Element 3', corpus: { id: 'corpus2' } }
        const page1 = {
          count: 4,
          previous: null,
          next: '/elements/selection/?page=2',
          number: 1,
          results: [element1, element2]
        }
        mock.onGet('/elements/selection/', { params: { page: 1 } }).reply(200, page1)
        const page2 = {
          count: 3,
          previous: '/elements/selection/?page=1',
          next: null,
          number: 2,
          // Element 2 is duplicated
          results: [element2, element3]
        }
        mock.onGet('/elements/selection/', { params: { page: 2 } }).reply(200, page2)

        await store.dispatch('selection/get', {})

        assert.strictEqual(mock.history.get.length, 2)
        assert.deepStrictEqual(mock.history.get[0].params, { page: 1 })
        assert.deepStrictEqual(mock.history.get[1].params, { page: 2 })

        assert.deepStrictEqual(store.history, [
          { action: 'selection/get', payload: {} },
          { mutation: 'selection/reset' },
          { mutation: 'elements/bulkSet', payload: [element1, element2] },
          { mutation: 'selection/set', payload: page1 },
          { mutation: 'elements/bulkSet', payload: [element2, element3] },
          { mutation: 'selection/set', payload: page2 }
        ])

        assert.deepStrictEqual(store.state.selection, {
          count: 3,
          selection: {
            corpus1: [
              'element1',
              'element2'
            ],
            corpus2: [
              'element3'
            ]
          }
        })
      })

      it('handles errors', async () => {
        mock.onGet('/elements/selection/').reply(500)
        await assertRejects(async () => { await store.dispatch('selection/get', {}) })
        assert.deepStrictEqual(store.history, [
          { action: 'selection/get', payload: {} },
          { mutation: 'selection/reset' },
          { mutation: 'selection/reset' },
          {
            mutation: 'notifications/notify',
            payload: {
              type: 'error',
              text: 'Request failed with status code 500'
            }
          }
        ])
      })
    })
    describe('select', () => {
      it('selects one or more elements', async () => {
        const element1 = { id: 'element1', name: 'Element 1', corpus: { id: 'corpus1' } }
        const element2 = { id: 'element2', name: 'Element 2', corpus: { id: 'corpus1' } }
        mock.onPost('/elements/selection/').reply(201)

        await store.dispatch('selection/select', { elements: [element1, element2] })

        assert.deepStrictEqual(store.history, [
          { action: 'selection/select', payload: { elements: [element1, element2] } },
          { mutation: 'elements/bulkSet', payload: [element1, element2] },
          { mutation: 'selection/set', payload: { results: [element1, element2] } }
        ])

        assert.strictEqual(mock.history.post.length, 1)
        assert.deepStrictEqual(mock.history.post[0].data, {
          ids: ['element1', 'element2']
        })
      })

      it('handles errors', async () => {
        mock.onPost('/elements/selection/').reply(500)
        await assertRejects(async () => { await store.dispatch('selection/select', { elements: ['element1'] }) })
        assert.deepStrictEqual(store.history, [
          { action: 'selection/select', payload: { elements: ['element1'] } },
          { mutation: 'selection/reset' },
          {
            mutation: 'notifications/notify',
            payload: {
              type: 'error',
              text: 'Request failed with status code 500'
            }
          }
        ])
      })
    })
    describe('unselect', () => {
      it('unselects an element', async () => {
        const element1 = { id: 'element1', name: 'Element 1', corpus: { id: 'corpus1' } }
        mock.onDelete('/elements/selection/').reply(204)
        store.setState('selection', {
          count: 3,
          selection: {
            corpus1: [
              'element1',
              'element2'
            ],
            corpus2: [
              'element3'
            ]
          }
        })

        await store.dispatch('selection/unselect', element1)

        assert.deepStrictEqual(store.history, [
          { action: 'selection/unselect', payload: element1 },
          { mutation: 'selection/remove', payload: element1 }
        ])

        assert.strictEqual(mock.history.delete.length, 1)
        assert.deepStrictEqual(mock.history.delete[0].data, { id: 'element1' })

        assert.deepStrictEqual(store.state.selection, {
          count: 2,
          selection: {
            corpus1: ['element2'],
            corpus2: ['element3']
          }
        })
      })

      it('handles errors', async () => {
        mock.onDelete('/elements/selection/').reply(500)
        await assertRejects(async () => { await store.dispatch('selection/unselect', { id: 'element1' }) })
        assert.deepStrictEqual(store.history, [
          { action: 'selection/unselect', payload: { id: 'element1' } },
          { mutation: 'selection/reset' },
          {
            mutation: 'notifications/notify',
            payload: {
              type: 'error',
              text: 'Request failed with status code 500'
            }
          }
        ])
      })
    })
    describe('clear', () => {
      it('clears the selection', async () => {
        mock.onDelete('/elements/selection/').reply(204)
        await store.dispatch('selection/clear')
        assert.deepStrictEqual(store.history, [
          { action: 'selection/clear' },
          { mutation: 'selection/reset' }
        ])
        assert.strictEqual(mock.history.delete.length, 1)
        assert.strictEqual(mock.history.delete[0].data, undefined)
      })

      it('handles errors', async () => {
        mock.onDelete('/elements/selection/').reply(500)
        await assertRejects(async () => { await store.dispatch('selection/clear') })
        assert.deepStrictEqual(store.history, [
          { action: 'selection/clear' },
          {
            mutation: 'notifications/notify',
            payload: {
              type: 'error',
              text: 'Request failed with status code 500'
            }
          },
          { mutation: 'selection/reset' }
        ])
      })
    })
    describe('validateClassifications', () => {
      it('validates classifications', async () => {
        mock.onPost('/classifications/selection/').reply(201)
        await store.dispatch('selection/validateClassifications', { corpusId: 'corpus1' })
        assert.deepStrictEqual(store.history, [
          { action: 'selection/validateClassifications', payload: { corpusId: 'corpus1' } }
        ])
        assert.strictEqual(mock.history.post.length, 1)
        assert.deepStrictEqual(mock.history.post[0].data, {
          mode: 'validate',
          corpus_id: 'corpus1'
        })
      })

      it('handles errors', async () => {
        mock.onPost('/classifications/selection/').reply(500)
        await assertRejects(async () => {
          await store.dispatch('selection/validateClassifications', { corpusId: 'corpus1' })
        })
        assert.deepStrictEqual(store.history, [
          { action: 'selection/validateClassifications', payload: { corpusId: 'corpus1' } },
          {
            mutation: 'notifications/notify',
            payload: {
              type: 'error',
              text: 'Request failed with status code 500'
            }
          }
        ])
      })
    })
    describe('createClassifications', () => {
      it('creates classifications', async () => {
        mock.onPost('/classifications/selection/').reply(201)
        await store.dispatch('selection/createClassifications', {
          corpusId: 'corpus1',
          mlClassId: 'something'
        })
        assert.deepStrictEqual(store.history, [
          {
            action: 'selection/createClassifications',
            payload: {
              corpusId: 'corpus1',
              mlClassId: 'something'
            }
          }
        ])
        assert.strictEqual(mock.history.post.length, 1)
        assert.deepStrictEqual(mock.history.post[0].data, {
          mode: 'create',
          corpus_id: 'corpus1',
          ml_class: 'something'
        })
      })

      it('handles errors', async () => {
        mock.onPost('/classifications/selection/').reply(500)
        await assertRejects(async () => {
          await store.dispatch('selection/createClassifications', { corpusId: 'corpus1', mlClassId: 'something' })
        })
        assert.deepStrictEqual(store.history, [
          {
            action: 'selection/createClassifications',
            payload: {
              corpusId: 'corpus1',
              mlClassId: 'something'
            }
          },
          {
            mutation: 'notifications/notify',
            payload: {
              type: 'error',
              text: 'Request failed with status code 500'
            }
          }
        ])
      })
    })
    describe('deleteElements', () => {
      it('delete selected elements', async () => {
        mock.onDelete('/corpus/corpus1/selection/').reply(204)
        mock.onGet('/jobs/').reply(200, jobsSample)
        store.state.selection = {
          count: 3,
          selection: {
            corpus1: [
              'element1',
              'element2'
            ],
            corpus2: [
              'element3'
            ]
          }
        }

        await store.dispatch('selection/deleteElements', { corpusId: 'corpus1' })
        await store.actionsCompleted()

        assert.deepStrictEqual(store.history, [
          { action: 'selection/deleteElements', payload: { corpusId: 'corpus1' } },
          { mutation: 'notifications/notify', payload: { type: 'success', text: 'Elements deletion has been scheduled.' } },
          { mutation: 'selection/removeCorpus', payload: 'corpus1' }
        ])
        assert.strictEqual(mock.history.delete.length, 1)
        assert.deepStrictEqual(mock.history.delete[0].data, undefined)

        assert.deepStrictEqual(store.state.selection, {
          count: 1,
          selection: {
            corpus2: ['element3']
          }
        })
        assert.deepStrictEqual(jobsStore.jobs, {
          jobid: jobsSample[0]
        })
      })

      it('handles errors', async () => {
        mock.onDelete('/corpus/corpus1/selection/').reply(500)
        mock.onGet('/jobs/').reply(200, jobsSample)

        await store.dispatch('selection/deleteElements', { corpusId: 'corpus1' })
        await store.actionsCompleted()

        assert.deepStrictEqual(store.history, [
          { action: 'selection/deleteElements', payload: { corpusId: 'corpus1' } },
          { mutation: 'notifications/notify', payload: { type: 'error', text: 'Request failed with status code 500' } }
        ])
        assert.deepStrictEqual(jobsStore.jobs, {
          jobid: jobsSample[0]
        })
      })
    })

    describe('moveSelection', () => {
      it('moves selected elements to a destination folder', async () => {
        mock.onGet('/jobs/').reply(200, jobsSample)
        mock.onPost('/elements/selection/move/').reply(201)

        await store.dispatch('selection/moveSelection', {
          corpusId: 'corpus1',
          destination: 'folder1'
        })
        await store.actionsCompleted()

        assert.deepStrictEqual(store.history, [
          {
            action: 'selection/moveSelection',
            payload: {
              corpusId: 'corpus1',
              destination: 'folder1'
            }
          },
          {
            mutation: 'notifications/notify',
            payload: {
              text: 'Elements moving has been scheduled.',
              type: 'success'
            }
          }
        ])
        assert.strictEqual(mock.history.post.length, 1)
        assert.deepStrictEqual(mock.history.post[0].data, {
          corpus_id: 'corpus1',
          destination: 'folder1'
        })

        assert.deepStrictEqual(jobsStore.jobs, {
          jobid: jobsSample[0]
        })
      })

      describe('handles errors', () => {
        const cases = [
          [
            'on the corpus ID',
            400,
            { corpus_id: ['No elements have been selected in this corpus.'] },
            'No elements have been selected in this corpus.'
          ],
          [
            'on the destination folder',
            400,
            { destination: ['An element cannot be moved into itself.'] },
            'An element cannot be moved into itself.'
          ],
          [
            'on feature flags',
            400,
            { detail: ['Selection is not available on this instance.'] },
            'Selection is not available on this instance.'
          ],
          [
            'with an unsupported error message',
            400,
            { blub: ['blob'] },
            'Request failed with status code 400'
          ],
          [
            'with no error message',
            418,
            null,
            'Request failed with status code 418'
          ]
        ]

        for (const [name, statusCode, response, expected] of cases) {
          it(name, async () => {
            mock.onGet('/jobs/').reply(200, jobsSample)
            mock.onPost('/elements/selection/move/').reply(statusCode, response)

            await store.dispatch('selection/moveSelection', {
              corpusId: 'corpus1',
              destination: 'folder1'
            })
            await store.actionsCompleted()

            assert.deepStrictEqual(store.history, [
              {
                action: 'selection/moveSelection',
                payload: {
                  corpusId: 'corpus1',
                  destination: 'folder1'
                }
              },
              { mutation: 'notifications/notify', payload: { type: 'error', text: expected } }
            ])
            assert.deepStrictEqual(jobsStore.jobs, {
              jobid: jobsSample[0]
            })
          })
        }
      })
    })

    describe('createParent', () => {
      it('adds selected elements to a destination folder', async () => {
        mock.onGet('/jobs/').reply(200, jobsSample)
        mock.onPost('/elements/selection/parent/').reply(201)

        await store.dispatch('selection/createParent', {
          corpusId: 'corpus1',
          parentId: 'folder1'
        })
        await store.actionsCompleted()

        assert.deepStrictEqual(store.history, [
          {
            action: 'selection/createParent',
            payload: {
              corpusId: 'corpus1',
              parentId: 'folder1'
            }
          },
          {
            mutation: 'notifications/notify',
            payload: {
              text: 'Element linking has been scheduled.',
              type: 'success'
            }
          }
        ])
        assert.strictEqual(mock.history.post.length, 1)
        assert.deepStrictEqual(mock.history.post[0].data, {
          corpus_id: 'corpus1',
          parent_id: 'folder1'
        })

        assert.deepStrictEqual(jobsStore.jobs, {
          jobid: jobsSample[0]
        })
      })

      describe('handles errors', () => {
        const cases = [
          [
            'on the corpus ID',
            400,
            { corpus_id: ['No elements have been selected in this corpus.'] },
            'No elements have been selected in this corpus.'
          ],
          [
            'on the destination folder',
            400,
            { parent_id: ['An element cannot be added as a parent of itself.'] },
            'An element cannot be added as a parent of itself.'
          ],
          [
            'on feature flags',
            400,
            { detail: ['Selection is not available on this instance.'] },
            'Selection is not available on this instance.'
          ],
          [
            'with an unsupported error message',
            400,
            { blub: ['blob'] },
            'Request failed with status code 400'
          ],
          [
            'with no error message',
            418,
            null,
            'Request failed with status code 418'
          ]
        ]

        for (const [name, statusCode, response, expected] of cases) {
          it(name, async () => {
            mock.onGet('/jobs/').reply(200, jobsSample)
            mock.onPost('/elements/selection/parent/').reply(statusCode, response)

            await store.dispatch('selection/createParent', {
              corpusId: 'corpus1',
              parentId: 'folder1'
            })
            await store.actionsCompleted()

            assert.deepStrictEqual(store.history, [
              {
                action: 'selection/createParent',
                payload: {
                  corpusId: 'corpus1',
                  parentId: 'folder1'
                }
              },
              { mutation: 'notifications/notify', payload: { type: 'error', text: expected } }
            ])
            assert.deepStrictEqual(jobsStore.jobs, {
              jobid: jobsSample[0]
            })
          })
        }
      })
    })
  })

  describe('getters', () => {
    it('loaded', () => {
      const state = {
        count: 1000,
        selection: {
          corpus1: [
            'element1',
            'element2'
          ],
          corpus2: [
            'element3'
          ]
        }
      }
      assert.strictEqual(getters.loaded(state), 3)
    })
  })
})
